import threading
import time
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator, Iterator, Optional

from dagster._annotations import experimental
from dagster._core.pipes.client import PipesMessageReader
from dagster._core.pipes.context import PipesMessageHandler
from dagster._core.pipes.utils import (
    _join_thread,
    extract_message_or_forward_to_stdout,
)
from dagster_pipes import PipesDefaultMessageWriter, PipesParams

if TYPE_CHECKING:
    from ray.job_submission import JobSubmissionClient


@experimental
class PipesRayJobSubmissionClientMessageReader(PipesMessageReader):
    """
    Dagster Pipes message reader for recieving messages from a Ray job.
    Will extract Dagster events and forward the rest to stdout.
    """

    def __init__(self):
        self._handler: Optional["PipesMessageHandler"] = None
        self._thread: Optional[threading.Thread] = None
        self.session_closed = threading.Event()

    @contextmanager
    def read_messages(self, handler: "PipesMessageHandler") -> Iterator[PipesParams]:
        #  This method should start a thread to continuously read messages from some location
        self._handler = handler

        try:
            yield {PipesDefaultMessageWriter.STDIO_KEY: PipesDefaultMessageWriter.STDOUT}
        finally:
            self.terminate()

    @property
    def handler(self) -> "PipesMessageHandler":
        if self._handler is None:
            raise Exception("PipesMessageHandler is only available while reading messages in open_pipes_session")

        return self._handler

    def terminate(self) -> None:
        self.session_closed.set()

        if self._thread is not None:
            _join_thread(self._thread, "ray job logs tailing")

        self._handler = None

    def consume_job_logs(self, client: "JobSubmissionClient", job_id: str, blocking: bool = False) -> None:
        if blocking:
            handle_job_logs(handler=self.handler, client=client, job_id=job_id, session_closed=None)
        else:
            self._thread = threading.Thread(
                target=handle_job_logs,
                kwargs={
                    "handler": self.handler,
                    "client": client,
                    "job_id": job_id,
                    "session_closed": self.session_closed,
                },
                daemon=True,
            )
            self._thread.start()

    def no_messages_debug_text(self) -> str:
        return "Tried to read messages from a Ray job logs, but no messages were received."


def handle_job_logs(
    handler: PipesMessageHandler, client: "JobSubmissionClient", job_id: str, session_closed: Optional[threading.Event]
):
    import asyncio

    async_tailer = client.tail_job_logs(job_id=job_id)

    # Backward compatible sync generator
    def tail_logs() -> Generator[str, None, None]:
        while True:
            try:
                try:
                    loop = asyncio.get_event_loop()
                except RuntimeError:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                yield loop.run_until_complete(async_tailer.__anext__())  # type: ignore
            except (StopAsyncIteration, ValueError):
                break

    session_closed_at = None

    for chunk in tail_logs():
        for log_line in chunk.split("\n"):
            if log_line:
                extract_message_or_forward_to_stdout(handler, log_line)

        if session_closed is not None and session_closed.is_set():
            if session_closed_at is None:
                session_closed_at = time.time()

            if time.time() - session_closed_at > 30:  # wait for 30 seconds to flush all logs
                session_closed.wait()
                return

    if session_closed is not None:
        session_closed.wait()

    return



from dagster_ray._base.resources import BaseRayResource
from dagster_ray.executor import ray_executor
from dagster_ray.io_manager import RayIOManager
from dagster_ray.run_launcher import RayRunLauncher

RayResource = BaseRayResource


__all__ = ["RayResource", "RayRunLauncher", "RayIOManager", "ray_executor"]



import os
import socket
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Generator

import pytest
import pytest_cases
from pytest_kubernetes.options import ClusterOptions
from pytest_kubernetes.providers import AClusterManager, select_provider_manager

from tests import ROOT_DIR


def get_random_free_port():
    sock = socket.socket()
    sock.bind(("", 0))
    return sock.getsockname()[1]


@pytest.fixture(scope="session")
def kuberay_helm_repo():
    subprocess.run(["helm", "repo", "add", "kuberay", "https://ray-project.github.io/kuberay-helm/"], check=True)
    subprocess.run(["helm", "repo", "update", "kuberay"], check=True)


PYTEST_DAGSTER_RAY_IMAGE = os.getenv("PYTEST_DAGSTER_RAY_IMAGE")


@pytest.fixture(scope="session")
def dagster_ray_image():
    """
    Either returns the image name from the environment variable PYTEST_DAGSTER_RAY_IMAGE
    or builds the image and returns it
    """

    if PYTEST_DAGSTER_RAY_IMAGE is None:
        # build the local image
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        image = f"local/dagster-ray:py-{python_version}"
        subprocess.run(
            [
                "docker",
                "build",
                "-f",
                str(ROOT_DIR / "Dockerfile"),
                "--build-arg",
                "BUILD_DEPENDENCIES=dev",
                "--build-arg",
                f"PYTHON_VERSION={python_version}",
                "-t",
                image,
                str(ROOT_DIR),
            ],
            env={
                "DOCKER_BUILDKIT": "1",
            },
            check=True,
        )
    else:
        image = PYTEST_DAGSTER_RAY_IMAGE

    return image


# TODO: it's easy to parametrize over different versions of k8s
# but it would take quite some time to test all of them!
# probably should only do it in CI
KUBERNETES_VERSION = os.getenv("PYTEST_KUBERNETES_VERSION", "1.31.0")

KUBERAY_VERSIONS = os.getenv("PYTEST_KUBERAY_VERSIONS", "1.2.2").split(",")

NAMESPACE = "ray"


@pytest_cases.fixture(scope="session")
@pytest.mark.parametrize("kuberay_version", KUBERAY_VERSIONS)
def k8s_with_kuberay(
    request, kuberay_helm_repo, dagster_ray_image: str, kuberay_version: str
) -> Generator[AClusterManager, None, None]:
    k8s = select_provider_manager("minikube")("dagster-ray")
    k8s.create(ClusterOptions(api_version=KUBERNETES_VERSION))
    # load images in advance to avoid possible timeouts later on
    k8s.load_image(f"quay.io/kuberay/operator:v{kuberay_version}")

    # warning: minikube fails to load the image directly because of
    # https://github.com/kubernetes/minikube/issues/18021
    # so we export it to .tar first
    # TODO: load image without .tar export once the issue is resolved
    with tempfile.TemporaryDirectory() as tmpdir:
        image_tar = Path(tmpdir) / "dagster-ray.tar"
        subprocess.run(["docker", "image", "save", "-o", str(image_tar), dagster_ray_image], check=True)
        k8s.load_image(str(image_tar))

    # init the cluster with a workload
    subprocess.run(
        [
            "helm",
            "--kubeconfig",
            str(k8s.kubeconfig),
            "upgrade",
            "--install",
            "--create-namespace",
            "--namespace",
            "kuberay-operator",
            "kuberay-operator",
            "kuberay/kuberay-operator",
            "--version",
            kuberay_version,
        ],
        check=True,
    )

    k8s.wait("deployment/kuberay-operator", "condition=Available=True", namespace="kuberay-operator")
    # namespace to create RayClusters in
    try:
        k8s.kubectl(["create", "namespace", NAMESPACE])
    except RuntimeError as e:
        if "AlreadyExists" not in str(e):
            raise

    yield k8s
    k8s.delete()
